#line 1454 "ExpandModules.nw"
#ifndef _DISMANTLE_HPP
#  define _DISMANTLE_HPP

#  include "AST.hpp"
#  include <assert.h>
#  include <sstream>
#  include <set>
#  include <map>

namespace ExpandModules {
  using namespace IR;
  using namespace AST;
  using std::set;
  using std::map;

  
#line 123 "ExpandModules.nw"
class FindRoots : public Visitor {
  Modules *modules;
public:
  set<Module*> roots;
  
#line 132 "ExpandModules.nw"
FindRoots(Modules *m) {
  assert(m);
  modules = m;
}

virtual ~FindRoots() {}
#line 146 "ExpandModules.nw"
const set<Module*> *find();
#line 172 "ExpandModules.nw"
void recurse(ASTNode* n) { if (n) n->welcome(*this); }
#line 182 "ExpandModules.nw"
Status visit(Run &);
#line 206 "ExpandModules.nw"
Status visit(Nothing &) { return Status(); }
Status visit(Pause &) { return Status(); }
Status visit(Halt &) { return Status(); }
Status visit(Exit &) { return Status(); }
Status visit(Emit &) { return Status(); }
Status visit(Sustain &) { return Status(); }
Status visit(Assign &) { return Status(); }
Status visit(ProcedureCall &) { return Status(); }
#line 223 "ExpandModules.nw"
Status visit(Module &m) { recurse(m.body); return Status(); }
Status visit(Loop &s) { recurse(s.body); return Status(); }
Status visit(Repeat &s) { recurse(s.body); return Status(); }
Status visit(Signal &s) { recurse(s.body); return Status(); }
Status visit(Var &s) { recurse(s.body); return Status(); }
Status visit(LoopEach &s) { recurse(s.body); return Status(); }
Status visit(Every &s) { recurse(s.body); return Status(); }
Status visit(Suspend &s) { recurse(s.body); return Status(); }
Status visit(DoUpto &s) { recurse(s.body); return Status(); }
#line 235 "ExpandModules.nw"
Status visit(DoWatching &s) {
  recurse(s.body);
  recurse(s.timeout);
  return Status();
}
#line 243 "ExpandModules.nw"
Status visit(Trap &s) {
  recurse(s.body);
  for (vector<PredicatedStatement*>::const_iterator i = s.handlers.begin() ;
       i != s.handlers.end() ; i++ ) {
    assert(*i);   
    recurse((*i)->body);
  }
  return Status();
}
#line 257 "ExpandModules.nw"
Status visit(StatementList &l) {
  for (vector<Statement*>::iterator i = l.statements.begin() ;
       i != l.statements.end() ; i++ )
    recurse(*i);
  return Status();
}
#line 266 "ExpandModules.nw"
Status visit(ParallelStatementList &l) {
  for (vector<Statement*>::iterator i = l.threads.begin() ;
       i != l.threads.end() ; i++ )
    recurse(*i);
  return Status();
}
#line 277 "ExpandModules.nw"
Status visit(IfThenElse &s) {
  recurse(s.then_part);
  recurse(s.else_part);
  return Status();
}
#line 285 "ExpandModules.nw"
Status visit(Exec &s) {
  for (vector<TaskCall*>::iterator i = s.calls.begin() ;
       i != s.calls.end() ; i++ ) {
    assert(*i);    
    recurse((*i)->body);
  }
  return Status();
}
#line 301 "ExpandModules.nw"
void visitCase(CaseStatement *s) {
  for ( vector<PredicatedStatement*>::const_iterator i = s->cases.begin() ;
	i != s->cases.end() ; i++ ) {
    assert(*i);
    recurse((*i)->body);
  }
  recurse(s->default_stmt);
}
#line 312 "ExpandModules.nw"
Status visit(Present &s) { visitCase( (CaseStatement*) &s); return Status(); }
Status visit(If &s) { visitCase( (CaseStatement*) &s); return Status(); }
Status visit(Await &s) { visitCase( (CaseStatement*) &s); return Status(); }
#line 318 "ExpandModules.nw"
Status visit(Abort &s) {
  visitCase( (CaseStatement*) &s);
  recurse(s.body);
  return Status();
}
#line 128 "ExpandModules.nw"
};
#line 1470 "ExpandModules.nw"
  class Copier;
  
#line 1377 "ExpandModules.nw"
class ExpressionCopier {
  Expression *theExpr;
  Copier *theCopier;
public:
  ExpressionCopier(Expression*, Copier*);
  Expression *copy();
};
#line 1472 "ExpandModules.nw"
  
#line 345 "ExpandModules.nw"
class Copier : public Visitor {
  map<const Symbol*, Symbol*> copiedSymbolMap;
  map<const SymbolTable*, SymbolTable*> copiedSymbolTableMap;
  map<const ConstantSymbol*, ExpressionCopier*> newConstantExpression;
  Modules *oldModules;
  Modules *newModules;
  Module *newModule;
public:
  
#line 358 "ExpandModules.nw"
Copier(Modules *om, Modules *nm)
  : oldModules(om), newModules(nm), newModule(0) {
  assert(nm);
  assert(om);
}

Copier(Copier *c) {
  assert(c);
  oldModules = c->oldModules;
  newModules = c->newModules;
  newModule = c->newModule;
}

virtual ~Copier() {}
#line 379 "ExpandModules.nw"
template <class T> T* copy(T* n) {
  T* result = n ? dynamic_cast<T*>(n->welcome(*this).n) : 0;
  assert(result || !n);
  return result;
}
#line 393 "ExpandModules.nw"
template <class T> T* newSymbol(const T* s) {
  if (!s) return 0;
  // std::cerr << "newSymbol " << s->name << std::endl;
  map<const Symbol*, Symbol*>::const_iterator i = copiedSymbolMap.find(s);
  assert(i != copiedSymbolMap.end());        // Should be in the map
  assert((*i).second);                       // Should be non-NULL
  T* result = dynamic_cast<T*>((*i).second);
  assert(result);                            // Should be the same type
  return result;
}
#line 408 "ExpandModules.nw"
SymbolTable *newSymbolTable(const SymbolTable *s) {
  assert(s);
  map<const SymbolTable*, SymbolTable*>::const_iterator i =
     copiedSymbolTableMap.find(s);
  assert(i != copiedSymbolTableMap.end());
  SymbolTable *result = (*i).second;
  assert(result);
  return result;
}
#line 427 "ExpandModules.nw"
void copySymbolTable(const SymbolTable *, SymbolTable *);
#line 457 "ExpandModules.nw"
Module *copyModule(Module&);
#line 494 "ExpandModules.nw"
Status visit(Exclusion &e) {
  Exclusion *result = new Exclusion();
  for ( vector<SignalSymbol*>::const_iterator i = e.signals.begin() ;
	i != e.signals.end() ; i++ ) {
    assert(*i);
    result->signals.push_back(newSymbol(*i));
  }
  return result;
}
#line 506 "ExpandModules.nw"
Status visit(Implication &e) {
  return new Implication(newSymbol(e.predicate), newSymbol(e.implication));
}
#line 519 "ExpandModules.nw"
Status visit(Run &);
#line 996 "ExpandModules.nw"
Status visit(TypeSymbol &s) { return new TypeSymbol(s.name); }

Status visit(BuiltinTypeSymbol &s) { return new BuiltinTypeSymbol(s.name); }

Status visit(ConstantSymbol &s) {
  return new ConstantSymbol(s.name, newSymbol(s.type), copy(s.initializer));
}

Status visit(BuiltinConstantSymbol &s) {
  return new BuiltinConstantSymbol(s.name, newSymbol(s.type), 
                                   copy(s.initializer));
}
#line 1011 "ExpandModules.nw"
Status visit(FunctionSymbol&);
Status visit(BuiltinFunctionSymbol&);
Status visit(ProcedureSymbol&);
Status visit(TaskSymbol&);
#line 1079 "ExpandModules.nw"
Status visit(SignalSymbol &s)
{
  // std::cout << "Copy signal " << s.name << std::endl;
  return
    new SignalSymbol( s.name, newSymbol(s.type), (SignalSymbol::kinds) s.kind,
		      newSymbol(s.combine), copy(s.initializer),
		      newSymbol(s.presence), newSymbol(s.value) );
}

Status visit(BuiltinSignalSymbol &s)
{
  return
    new BuiltinSignalSymbol( s.name, newSymbol(s.type),
	                     (SignalSymbol::kinds) s.kind,
			     newSymbol(s.combine), newSymbol(s.presence),
			     newSymbol(s.value) );
}
#line 1099 "ExpandModules.nw"
Status visit(VariableSymbol &s) {
  return new VariableSymbol(s.name, newSymbol(s.type), copy(s.initializer));
}
#line 1109 "ExpandModules.nw"
Status visit(Nothing &) { return new Nothing(); }
Status visit(Pause &) { return new Pause(); }
Status visit(Halt &) { return new Halt(); }
#line 1117 "ExpandModules.nw"
Status visit(Emit &e) {
  return new Emit(newSymbol(e.signal), copy(e.value));
}

Status visit(Sustain &s) {
  return new Sustain(newSymbol(s.signal), copy(s.value));
}
#line 1129 "ExpandModules.nw"
Status visit(Assign &a) {
  return new Assign(newSymbol(a.variable), copy(a.value));
}
#line 1135 "ExpandModules.nw"
Status visit(Exit &e) { return new Exit(newSymbol(e.trap), copy(e.value)); }
#line 1139 "ExpandModules.nw"
Status visit(ProcedureCall &);
#line 1165 "ExpandModules.nw"
Status visit(StatementList&);
Status visit(ParallelStatementList&);
#line 1194 "ExpandModules.nw"
Status visit(IfThenElse &s) {
  return new IfThenElse(copy(s.predicate), copy(s.then_part),
			copy(s.else_part));
}
#line 1201 "ExpandModules.nw"
Status visit(Loop &s) { return new Loop(copy(s.body)); }
#line 1205 "ExpandModules.nw"
Status visit(Repeat &s) {
  return new Repeat(copy(s.body), copy(s.count), s.is_positive);
}
#line 1211 "ExpandModules.nw"
Status visit(LoopEach &s) {
  return new LoopEach( copy(s.body), copy(s.predicate));
}
#line 1217 "ExpandModules.nw"
Status visit(Every &s) {
  return new Every( copy(s.body), copy(s.predicate));
}
#line 1223 "ExpandModules.nw"
Status visit(Suspend &s) {
  return new Suspend( copy(s.body), copy(s.predicate));
}
#line 1229 "ExpandModules.nw"
Status visit(DoWatching &s) {
  return new DoWatching( copy(s.body), copy(s.predicate), copy(s.timeout));
}
#line 1235 "ExpandModules.nw"
Status visit(DoUpto &s) {
  return new DoUpto( copy(s.body), copy(s.predicate));
}
#line 1241 "ExpandModules.nw"
Status visit(Exec &);
#line 1277 "ExpandModules.nw"
void copyCases(const CaseStatement &, CaseStatement *);
#line 1295 "ExpandModules.nw"
Status visit(Abort &s) {
  Abort *result = new Abort(copy(s.body), s.is_weak);
  copyCases(s, result);
  return result;
}
#line 1303 "ExpandModules.nw"
Status visit(Await &s) {
  Await *result = new Await();
  copyCases(s, result);
  return result;
}
#line 1311 "ExpandModules.nw"
Status visit(Present &s) {
  Present *result = new Present();
  copyCases(s, result);
  return result;
}

Status visit(If &s) {
  If *result = new If();
  copyCases(s, result);
  return result;
}
#line 1327 "ExpandModules.nw"
Status visit(Var &s) {
  Var *result = new Var();
  result->symbols = new SymbolTable();
  copySymbolTable(s.symbols, result->symbols);
  result->body = copy(s.body);
  return result;
}
#line 1337 "ExpandModules.nw"
Status visit(Signal &s) {
  Signal *result = new Signal();
  result->symbols = new SymbolTable();
  copySymbolTable(s.symbols, result->symbols);
  result->body = copy(s.body);
  return result;
}
#line 1347 "ExpandModules.nw"
Status visit(Trap &s) {
  Trap *result = new Trap();
  result->symbols = new SymbolTable();
  copySymbolTable(s.symbols, result->symbols);
  result->body = copy(s.body);
  for ( vector<PredicatedStatement *>::const_iterator i = s.handlers.begin() ;
        i != s.handlers.end() ; i++ ) {
    assert(*i);
    result->newHandler( copy((*i)->predicate), copy((*i)->body) );
  }
  return result;
}
#line 1367 "ExpandModules.nw"
Status visit(Literal &l) { return new Literal(l.value, newSymbol(l.type)); }
#line 1394 "ExpandModules.nw"
Status visit(LoadVariableExpression &e) {
  ConstantSymbol *cs = dynamic_cast<ConstantSymbol*>(e.variable);
  if (cs) {
    map<const ConstantSymbol*, ExpressionCopier*>::iterator i =
  	newConstantExpression.find(cs);
    if ( i != newConstantExpression.end() )
      return (*i).second->copy();
  }
  return new LoadVariableExpression(newSymbol(e.variable));
}
#line 1407 "ExpandModules.nw"
Status visit(LoadSignalExpression &e) {
  return new LoadSignalExpression(e.type, newSymbol(e.signal));
}

Status visit(LoadSignalValueExpression &e) {
  return new LoadSignalValueExpression(newSymbol(e.signal));
}
#line 1417 "ExpandModules.nw"
Status visit(Delay &d) {
  return new Delay(newSymbol(d.type), copy(d.predicate),
		   copy(d.count), d.is_immediate);
}
#line 1424 "ExpandModules.nw"
Status visit(UnaryOp &o) {
  return new UnaryOp(newSymbol(o.type), o.op, copy(o.source));
}

Status visit(BinaryOp &o) {
  return new BinaryOp(newSymbol(o.type), o.op,
		      copy(o.source1), copy(o.source2));
}
#line 1435 "ExpandModules.nw"
Status visit(FunctionCall &);
#line 354 "ExpandModules.nw"
};
#line 1473 "ExpandModules.nw"
  
#line 90 "ExpandModules.nw"
Modules *rewrite(Modules *);
#line 1474 "ExpandModules.nw"
}
#endif
